.equ HEAPSIZE=1024

; reserve HEAPSIZE bytes for the heap

.dseg
heap_first_free: .byte 2
heap_start:
.byte heapsize
heap_end:

.cseg

;memory block
; 0 previous block 2 bytes
; 1
; 2 next block 2 bytes
; 3
; 4... data

.equ HEAP_DATAOFFSET = 4;


heap_init: ;sets the first block to 0 0
    LDI Zl, low(heap_first_free)
    LDI Zh,high(heap_first_free)
    LDI r16,low(heap_start)
    ST Z+,r16
    LDI r16,high(heap_start)
    ST Z,r16
    LDI Zl, low(heap_start)
    LDI Zh,high(heap_start)
    LDI r16,0
    ST Z+,r16
    ST Z+,r16
    ST Z+,r16
    ST Z+,r16
  ret
  

heap_sizeof: ;pointer in Y;result in r17:r16
    PUSH zl
    PUSH zh
    movw zh:zl,yh:yl
    SBIW Zh:zl,4
    LDD r17,Z+3
    LDD r16,Z+2
    SUB r16,Zl
    SBC r17,Zh
     
    POP zh
    POP zl
    ret      
    
    
heap_malloc: ;size in r17:r16, pointer in Y
    push r18
    push r24
    push r25
    push xl
    push xh
    push zl
    push zh
    ;r18 X Z
    SUBI r16, low(-HEAP_DATAOFFSET)
    SBCI r17,high(-HEAP_DATAOFFSET)
    LDS yl,heap_first_free
    LDS yh,heap_first_free+1
  heap_malloc_start:
    LDD xl,Y+2
    LDD xh,Y+3
    MOV r18,Xh
    CBR Xh,$80
    MOVW zh:zl,Yh:Yl
    ADD zl,r16
    ADC zh,r17
    CPI Xh,0
    CPC Xh,Xl
    BREQ heap_malloc_last_block
    SBRC r18,7 ;Bit 7 set => Free block
    RJMP heap_malloc_free_block
  heap_malloc_next_block:
    MOVW Yh:Yl,Xh:Xl
    rjmp heap_malloc_start
    
  heap_malloc_free_block:
    CP  xl,zl
    CPC xh,zh
    BRLT heap_malloc_next_block ; if to small
    
    MOVW r25:r24,zh:zl
    ADIW r25:r24,HEAP_DATAOFFSET ;minimal data needed for split
    CP  xl,r24
    CPC xh,r25
    BRGE heap_malloc_split_block
    ;take whole block without change
    STD Y+2,xl
    STD Y+3,xh
    rjmp heap_malloc_return_y    

;    rjmp heap_malloc_next_block   
    
  heap_malloc_last_block:
    LDI r18,high(heap_end)
    CPI Zl,  low(heap_end)
    CP Zh,r18
    BRGE heap_malloc_memory_full
    STD Z+0,Yl
    STD Z+1,Yh
    LDI r18,0
    STD Z+2,r18
    STD Z+3,r18 
    STD Y+2,Zl
    STD Y+3,Zh
    rjmp heap_malloc_return_y

  heap_malloc_memory_full:
    LDI yh,0
    LDI yl,0
    rjmp heap_malloc_return    
    
  heap_malloc_return_y:
    ADIW yh:yl,HEAP_DATAOFFSET

  heap_malloc_return:    
    pop zh
    pop zl
    pop xh
    pop xl
    pop r25
    pop r24
    pop r18
    ret

  heap_malloc_split_block: ;Probably the most dangerous piece here
    ;Y points to current block
    ;X points to next block
    ;Z points to where the middle block would be
    
    STD Z+0,Yl ;set middle.previous to current
    STD Z+1,Yh 
    STD Z+2,Xl ;set middle.next to next
    MOV r25,Xh
    SBR r25,$80 ;set empty bit
    STD Z+3,r25
    
    STD Y+2,Zl ;set current.next to middle
    STD Y+3,Zh

    ST X+,Zl ;set next.previous to middle
    ST X+,Zh
    
    rjmp heap_malloc_return_y

heap_free: ;y is the pointer of the block to free
    CPI Yl,0
    CPC Yl,Yl
    breq heap_free_end
    ;r17:r16
    push r16
    push r17
    push Zl
    push zh
    SUBI Yl, low(HEAP_DATAOFFSET)
    SBCI Yh,high(HEAP_DATAOFFSET)
    LDD r17,Y+3
    SBR r17,$80
    STD Y+3,r17
    
  heap_free_cleanup_forward:    
    LDD Zl,Y+2
    LDD Zh,Y+3
    CBR Zh,$80
    LDD r16,Z+2
    LDD r17,Z+3
    SBRS r17,7 ;is the next block free
    rjmp heap_free_cleanup_forward_end
    CPI r16,0 ;is last block ?
    CPC r16,r17
    BREQ heap_free_cleanup_forward_end
    MOVW zh:zl,r17:r16 ;set previous of new next block
    STD Y+2,r16
    SBR r17,$80
    STD Y+3,r17
    STD Z+0,Yl
    STD Z+1,Yh
    rjmp heap_free_cleanup_forward    
  heap_free_cleanup_forward_end:
  
  heap_free_cleanup_backward:
    
  heap_free_cleanup_backward_end:    
    pop zh
    pop zl
    pop r17
    pop r16
  heap_free_end:    
    ret    
    
