;routines for serial port

serial_init:
    push r16
    push r17
    LDI r16, low(103) ;103=9600baud
    LDI r17,high(103)
    OUT ubrrh,r17
    OUT ubrrl,r16
    LDI r16, (1<<RXEN)|(1<<TXEN)
    OUT UCSRB,r16
    ;8 data 2 stop bit
    LDI r16,(1<<URSEL)|(1<<USBS)|(3<<UCSZ0)
    OUT UCSRC,r16
    pop r16
    pop r17
    ret
    

serial_output: ;sends the byte in r16 to the serial port
  serial_output_loop: ;wait till puffer free
    sbis UCSRA,UDRE
    rjmp serial_output_loop
    OUT udr,r16
    ret
    
    
serial_input: ;waits for byte from serial interface
  serial_input_loop: ;wait till there is a byte
    sbis UCSRA,RXC
    rjmp serial_input_loop
    in r16,UDR
    ret
    
    
serial_output_string_pm: ;writes zero-terminated string in programm memory
    push r16
  serial_write_pm_loop:
    LPM r16,Z+
    call serial_output
    CPI r16,0
    BRNE serial_write_pm_loop
    pop r16  
    ret
    
serial_output_16bit: ;writes a 16-bit number in r17:r16 to the serial out
    push r18
    mov r18,r16
;    CBR r17,7 ;unset sign flag
    LDI r16,'0'-1
  serial_write_16_bit_10000:
    inc r16
    SUBI r18, low(10000)
    SBCI r17,high(10000)    
    BRGE serial_write_16_bit_10000
    SUBI r18, low(-10000)
    SBCI r17,high(-10000)    
    call serial_output
    
    LDI r16,'0'-1
  serial_write_16_bit_1000:
    inc r16
    SUBI r18, low(1000)
    SBCI r17,high(1000)    
    BRGE serial_write_16_bit_1000
    SUBI r18, low(-1000)
    SBCI r17,high(-1000)    
    call serial_output
    
    LDI r16,'0'-1
  serial_write_16_bit_100:
    inc r16
    SUBI r18, low(100)
    SBCI r17,high(100)    
    BRGE serial_write_16_bit_100
    SUBI r18, low(-100)
    SBCI r17,high(-100)    
    call serial_output
    
    LDI r16,'0'-1
  serial_write_16_bit_10:
    inc r16
    SUBI r18, low(10)
    SBCI r17,high(10)    
    BRGE serial_write_16_bit_10
    SUBI r18, low(-10)
    SBCI r17,high(-10)
    call serial_output    
    
    LDI r16,'0'
    ADD r16,r18   
    CPI r16,'9'+1
    BRGE serial_write_16_bit_error
    CPI r16,'0'
    BRLT serial_write_16_bit_error
    call serial_output
  serial_write_16_bit_error_return:    
    pop r18   
    ret
    
  serial_write_16_bit_error:
    LDI r16,'X'
    call serial_output
    rjmp serial_write_16_bit_error_return       
    
serial_output_crlf:
    push r16
    LDI r16,13
    call serial_output
    LDI r16,10
    call serial_output
    pop r16    
    ret
    
hextable:
.db "0123456789ABCDEF"    
    
serial_output_hex:
    push r16
    push r17
    push zl
    push zh
    MOV r17,r16
    ANDI r17,$0f
    LDI Zl, low(hextable*2)
    LDI Zh,high(hextable*2)
    ADD Zl,r17
    LDI r17,0
    ADC Zh,r17
    MOV r17,r16
    LPM r16,Z
    call serial_output
    SWAP r17
    ANDI r17,$0f
    LDI Zl, low(hextable*2)
    LDI Zh,high(hextable*2)
    ADD Zl,r17
    LDI r17,0
    ADC Zh,r17
    LPM r16,Z
    call serial_output
    pop zh
    pop zl
    pop r17
    pop r16
    ret    
    
    